trait ConvertToString {
    /// Returns a string representation of self
    fn to_string(&self) -> String;
}

#[derive(Clone, Copy, Debug, PartialEq)]
enum Square {
    Empty,
    Cross,
    Circle,
}

impl ConvertToString for Square {
    fn to_string(&self) -> String {
        match self {
            Square::Empty => String::from("."),
            Square::Cross => String::from("X"),
            Square::Circle => String::from("O"),
        }
    }
}

struct Board {
    board: [Square; 9],
}

impl Board {
    /// Creates a new empty board
    fn new() -> Self {
        Self {
            board: [Square::Empty; 9],
        }
    }

    /// Display the board to standard output
    fn display(&self) {
        println!("{}", self.to_string());
    }
}

impl ConvertToString for Board {
    fn to_string(&self) -> String {
        let mut out = String::new();
        for row in 0..3 {
            out.push_str(
                format!(
                    " {} | {} | {}",
                    self.board[row * 3].to_string(),
                    self.board[row * 3 + 1].to_string(),
                    self.board[row * 3 + 2].to_string()
                )
                .as_str(),
            );
            if row == 0 || row == 1 {
                out.push_str("\n---+---+---\n");
            }
        }
        out
    }
}

fn main() {
    let board = Board::new();

    board.display();
}

// Will only be build in 'test' configuration
#[cfg(test)]
// A module where the tests are located, more on modules later
mod tests {
    use super::*;

    #[test]
    fn should_pass() {
        assert!(true);
    }
    #[test]
    #[should_panic]
    fn should_panic() {
        panic!("Doing wrong things on purpose here!");
    }

    #[test]
    fn new_board_squares_are_empty() {
        let board = Board::new();

        for square in board.board {
            assert_eq!(square, Square::Empty);
        }
    }

    #[test]
    fn empty_board_to_string() {
        let expected = " . | . | .
---+---+---
 . | . | .
---+---+---
 . | . | .";
        assert_eq!(expected, Board::new().to_string());
    }

    #[test]
    fn empty_square_to_string_dot() {
        assert_eq!(".", Square::Empty.to_string());
    }
    #[test]
    fn cross_square_to_string_x() {
        assert_eq!("X", Square::Cross.to_string());
    }
    #[test]
    fn circle_square_to_string_big_o() {
        assert_eq!("O", Square::Circle.to_string());
    }
}
